import re
import warnings

import numpy as np
from monty.json import MSONable, jsanitize

from pydirac.core.orbitals import AtomicOrbital, MoleculeOrbitals, OrbitalType
from pydirac.io.inputs import Inp, Mol

__all__ = ["Output"]


class Output:
    """Class to parse DIRAC output file

    Attributes:
        filename (str): output filename of DIRAC. Normally, a good output file
            contains `inp` file and `mol` file, thus one can parse them here
            together.

    """

    def __init__(self, filename):
        self.filename = filename

        self.inp = None
        self.mol = None

        # these are attributes of Output
        self.is_ok = False
        self.energy_settings = {}

        self.parse_energy()

    def parse_energy(self):
        """Parse the whole output file

        Parse the whole output file including `inp` file and `mol` file and
        `orbit` information and calculation `results` in the `output` file.

        Note:
            Input:
                1) SO or SR,
                2) zff,
                3) dipole or quadrupole,
                4) CC or CI,
                5) X2C or DOSSSS
            Mol:
                1) molecule
                2) basis sets: default (e.g., dyall.acv4z) or explicit

        Returns:
            None
        """

        self.check_valid()

        if self.is_ok:
            self.parse_input()
            self.parse_mol()
            # self.parse_orbit()

            self.parse_results()

    def parse_input(self):
        """Parse input from output file

        Returns:
            input settings
        """

        start_line = re.compile(r"^Contents of the input file\s*")
        dash_line = re.compile(r"^-+")
        empty_line = re.compile(r"^$")
        comment_line = re.compile(r"^[#|!]+")
        inp_start_line = re.compile(r"^\*\*DIRAC\s*")
        endline = re.compile(r"^\s*\*END OF")

        with open(self.filename) as f:
            context = f.readlines()

        for i, line in enumerate(context):
            if start_line.match(line):
                context = context[i + 1 :]

        inp_str = []
        is_start = False
        for i, line in enumerate(context):
            if is_start and not endline.match(line):
                inp_str.append(line)
            elif inp_start_line.match(line):
                is_start = True
                inp_str.append(line)
            elif dash_line.match(line) or empty_line.match(line) or comment_line.match(line):
                continue
            else:
                inp_str.append(line)
                break

        self.inp = Inp.from_string(inp_str)

    def parse_mol(self):
        """Parse mol from output file

        Returns:
            Mol settings
        """
        start_line = re.compile(r"^Contents of the molecule file\s*")
        dash_line = re.compile(r"^-+")
        empty_line = re.compile(r"^$")
        # inp_start_line = re.compile(r"^\*\*DIRAC\s*")
        nuclei_nb_line = re.compile(
            r"\s*(?P<nuclei>[-+]?(\d+(\.\d*)?|\d*\.\d+))\s+(?P<nb_atoms>\d+)"
        )
        basis_line = re.compile(r"^\b(?:LARGE|EXPLICIT)\b\s+BASIS\s+(\S+)\s*")
        endline = re.compile(r"^FINISH")
        atomic_calc_ptn = re.compile(r"This is an atomic calculation")

        basis_type = None

        with open(self.filename) as f:
            context = f.readlines()

        if atomic_calc_ptn.search("".join(context), re.MULTILINE | re.DOTALL):
            is_atom = True
        else:
            is_atom = False

        mol_str = []
        for i, line in enumerate(context):
            if start_line.match(line):
                # context = context[i + 1:]
                context = context[i + 3 :]
                break
        for i, line in enumerate(context):
            if endline.match(line):
                break
            else:
                mol_str.append(line)

        self.mol = Mol.from_string(mol_str)
        self.mol.molecule.is_atom = is_atom

    def parse_results(self):
        """Parse calculation results from output file

        Returns:
            None
        """

        if not self.inp:
            self.parse_input()

        if not self.mol:
            self.parse_mol()

        # check if calculation is finished
        if self.inp.calc_method in ["CI", "CC"]:
            if self.inp.calc_method == "CC":
                self._parse_coupled_cluster()
            else:
                self._parse_configuration_interaction()
        else:
            self.energy_settings = {}

        calc_type = self.inp.calc_type or "null"
        hamiltonian = self.inp.hamiltonian or "null"
        calc_method = self.inp.calc_method or "null"
        basis_type = self.mol.basis_type or "null"

        # for a set of calculations, this task_type can be regarded as an id
        self._task_type = "-".join([calc_type, hamiltonian, calc_method]) + "@" + basis_type.strip()
        # print(self._task_type)

    def parse_orbit(self):
        """

        Args:
            filename:

            For SO
            * Fermion symmetry E1
                * Closed shell, f = 1.0000
                  -2.47797334037  ( 2)
                * Open shell #1, f = 0.5000
                  -0.13783115994  ( 2)
                * Virtual eigenvalues, f = 0.0000
                   0.00374548932  ( 6)        0.00499273090  ( 2)        0.01402796501  ( 6)        0.02179466435  (10)        0.03009547324  ( 2)
                   0.03716487091  ( 6)        0.06363549163  (14)        0.07606125973  (10)        0.09203286658  ( 6)        0.12208961664  ( 2)
                   0.19147871966  (14)        0.21707019780  (10)        0.22905557723  ( 2)        0.22905883825  ( 4)        0.26586221717  (18)
                   0.39444173531  ( 2)        0.49420939940  (14)        0.55623366579  ( 2)        0.55624525507  ( 4)        0.60447384566  ( 4)
                   0.60447617309  ( 6)        0.77688901369  (18)        1.25239748512  ( 2)        1.25997676379  ( 6)        1.25998182039  ( 8)
                   1.36133214856  ( 2)        1.36137529052  ( 4)        1.75072060544  ( 4)        1.75074035505  ( 6)        3.49702377155  ( 2)
                   3.49722182837  ( 4)        3.76715748401  ( 2)        9.76930432334  ( 2)        9.77034247125  ( 4)       10.47049017426  ( 2)
                  28.40952020368  ( 2)       36.12853032246  ( 2)       36.14097956915  ( 4)       79.15864898987  ( 2)      236.13953875683  ( 2)
                 783.60786051767  ( 2)     3002.96623263569  ( 2)    13731.73008778683  ( 2)

            For SR
            * Boson symmetry A1
              * Closed shell, f = 1.0000
                -2.47787107822  ( 2)
              * Open shell #1, f = 0.5000
                -0.13782962396  ( 2)
              * Virtual eigenvalues, f = 0.0000
                 0.01000493192  ( 2)        0.01881578996  ( 2)        0.03354206140  ( 2)        0.06105795349  ( 4)        0.08825402911  ( 2)
                 0.11027718518  ( 2)        0.15768768851  ( 4)        0.20553440208  ( 4)        0.22554311803  ( 2)        0.38377353600  ( 2)
                 0.46966416528  ( 4)        0.55315302017  ( 2)        0.59582793164  ( 4)        0.67809753506  ( 6)        1.24252600686  ( 2)
                 1.24308361482  ( 4)        1.35864972273  ( 2)        1.74463236465  ( 4)        3.49480571226  ( 2)        3.75812114959  ( 2)
                 9.76801822932  ( 2)       10.46222624083  ( 2)       28.40185035569  ( 2)       36.13522290609  ( 2)       79.15135049148  ( 2)
               236.13236364858  ( 2)      783.60069042839  ( 2)     3002.95967127327  ( 2)    13731.72642240539  ( 2)

            * Boson symmetry B1
              * Virtual eigenvalues, f = 0.0000
                 0.01000493192  ( 2)        0.03354206140  ( 2)        0.06105795349  ( 2)        0.08825402911  ( 2)        0.15768768851  ( 4)
                 0.20553440208  ( 2)        0.22554311803  ( 2)        0.46966416528  ( 4)        0.55315302017  ( 2)        0.59582793164  ( 2)
                 0.67809753506  ( 4)        1.24308361482  ( 4)        1.35864972273  ( 2)        1.74463236465  ( 2)        3.49480571226  ( 2)
                 9.76801822932  ( 2)       36.13522290609  ( 2)

        Returns:

        """
        start_ptn = re.compile(r"\s+Eigenvalues\s+")
        sym_ptn = re.compile(r"^\* \b(?:Boson|Fermion)\b symmetry (.*)")
        open_ptn = re.compile(r"^\s+\*\s+Open shell #\d+, f = (\d+(\.\d*)?)")
        closed_ptn = re.compile(r"^\s+\*\s+Closed shell, f = (\d+(\.\d*)?)")
        virtual_ptn = re.compile(r"\s+\*\s+Virtual eigenvalues, f = (\d+(\.\d*)?)")
        number = re.compile(r"(?P<num>[-+]?(\d+(\.\d*)?|\d*\.\d+))\s+\(\s*(?P<degen>\d+)\)")
        number_line_ptn = re.compile(r"\s+[-+]?(\d+(\.\d+)?|\d*\.\d+) .*")
        endline_ptn = re.compile(r"^\* HOMO - LUMO")

        with open(self.filename) as f:
            lines = f.readlines()

        for i, l in enumerate(lines):
            if start_ptn.match(l):
                lines = lines[i + 1 :]
                break
        else:
            warnings.warn(
                "No SCF calculation info founded, this is calculation"
                " without SCF, please check it carefully!"
            )

        cur_sym = None
        ao_type = None
        occ_frac = 0.00
        aos = []
        for l in lines:
            symmetry_match = sym_ptn.match(l)
            if symmetry_match:
                # add new branch
                cur_sym = symmetry_match.group(1)

            openshell_match = open_ptn.match(l)
            if openshell_match:
                occ_frac = float(openshell_match.group(1))
                ao_type = OrbitalType.OPEN_SHELL

            closeshell_match = closed_ptn.match(l)
            if closeshell_match:
                occ_frac = float(closeshell_match.group(1))
                # occ_frac = 1.0
                ao_type = OrbitalType.CLOSED_SHELL

            virtualshell_match = virtual_ptn.match(l)
            if virtualshell_match:
                occ_frac = float(virtualshell_match.group(1))
                # occ_frac = 0.0
                ao_type = OrbitalType.VIRTUAL_SHELL

            if number_line_ptn.match(l):
                ao_energy, ao_degen = [], []
                for m in number.finditer(l):
                    ao_energy.append(float(m.group("num")))
                    ao_degen.append(int(m.group("degen")))
                for e, d in zip(ao_energy, ao_degen):
                    ao = AtomicOrbital(cur_sym, ao_type, e, int(d), occ_frac)
                    aos.append(ao)

            if endline_ptn.match(l):
                break

        self.mos = MoleculeOrbitals(aos)
        return self.mos

    @property
    def task_type(self):
        """Get a task type from an output file

        Returns:
        """
        return self._task_type

    def check_valid(self):
        """
            ********** E N D   of   D I R A C  output  **********
        Returns:

        """
        end_pattern = re.compile(r"\s*\*+\s+E N D\s+of\s+D I R A C\s+output\s+\*+")
        with open(self.filename) as f:
            context = f.read()
        if end_pattern.search(context):
            self.is_ok = True
        else:
            self.is_ok = False
        return self.is_ok

    def _parse_coupled_cluster(self, has_T=True):
        """Deal with RELCCSD calculations

        Notes:
            An example with RELCCSD calculations

             Overview of calculated energies
            @ SCF energy :                             -5880.437843209574567
            @ MP2 correlation energy :                    -0.486037237589513
            @ CCSD correlation energy :                   -0.451464950662248
            @ 4th order triples correction :              -0.016640269726526
            @ 5th order triples (T) correction :           0.000318657059527
            @ 5th order triples -T  correction :           0.000362317692980
            @ Total MP2 energy :                       -5880.923880447164265
            @ Total CCSD energy :                      -5880.889308160236396
            @ Total CCSD+T  energy :                   -5880.905948429963246
            @ Total CCSD(T) energy :                   -5880.905629772903922
            @ Total CCSD-T  energy :                   -5880.905586112270612

        Returns:
            None
        """
        start_pattern = re.compile(r"^\s*Overview of calculated energies")
        scf_pattern = re.compile(r"^@ SCF energy :\s+(?P<energy>[-+]?(\d+(\.\d*)?|\d*\.\d+))")
        mp2_pattern = re.compile(r"^@ Total MP2 energy :\s+(?P<energy>[-+]?(\d+(\.\d*)?|\d*\.\d+))")
        ccsd_pattern = re.compile(
            r"^@ Total CCSD energy :\s+(?P<energy>[-+]?(\d+(\.\d*)?|\d*\.\d+))"
        )
        ccsd_p_T_pattern = re.compile(
            r"^@ Total CCSD\(T\) energy :\s+(?P<energy>[-+]?(\d+(\.\d*)?|\d*\.\d+))"
        )
        end_line = re.compile(r"^$")

        with open(self.filename) as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                if start_pattern.match(line):
                    lines = lines[i:]
                    break

        energies = {}
        if has_T:
            patterns = [scf_pattern, mp2_pattern, ccsd_pattern, ccsd_p_T_pattern]
            keys = ["scf_e", "mp2_e", "ccsd_e", "ccsd(t)_e"]
        else:
            patterns = [scf_pattern, mp2_pattern, ccsd_pattern]
            keys = ["scf_e", "mp2_e", "ccsd_e"]

        for i, line in enumerate(lines):
            if end_line.match(line):
                break

            for pattern, k in zip(patterns, keys):
                m = pattern.match(line)
                if m:
                    energies[k] = float(m.group("energy"))
                    break

        self.energy_settings = energies
        self._parse_cc_orbit()

    def _parse_cc_orbit(self):
        # search orbital info
        #
        #  Configuration in highest pointgroup
        #                                           E    E
        #  Spinor class : occupied                  13   11
        #  Spinor class : virtual                  220  222
        #
        #  Configuration in abelian subgroup
        orb_start_pattern = re.compile(r"^\s*Configuration in highest pointgroup")
        occ_pattern = re.compile(r"^\s*Spinor class : occupied\s+((?:\d+\s+)+)")
        vir_pattern = re.compile(r"^\s*Spinor class : virtual\s+((?:\d+\s+)+)")
        orb_end_pattern = re.compile(r"^\s*Configuration in abelian subgroup")

        with open(self.filename) as f:
            lines = f.readlines()
            for i, line in enumerate(lines):
                if orb_start_pattern.match(line):
                    lines = lines[i:]
                    break

        clc_orb = {}
        keys = ["occ", "vir"]
        patterns = [occ_pattern, vir_pattern]

        for i, line in enumerate(lines):
            if orb_end_pattern.match(line):
                break

            for pattern, k in zip(patterns, keys):
                m = pattern.match(line)
                if m:
                    clc_orb[k] = sum([int(nb) for nb in m.group(1).split()])
                    break
        self._cc_calc_orbit = clc_orb

    def _parse_configuration_interaction(self):
        """Deal with KRCI calculations

        Notes:
            CI example:
                &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
                &&& KRCI calculation for symmetry no.    3
                &&& Number of CI roots for this symmetry       3
                &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

                Final CI energies  =  -5880.8481422675641  -5880.8380283826846  -5880.8380141759062

                root    1 ...... converged!

                root    2 ...... converged!

                root    3 ...... converged!

        Returns:
            None
        """
        # start_line = re.compile(r'Final CI energies\s+=\s+(?P<energy>([-+]?(\d+(\.\d*)?|\d*\.\d+)\s+)+)\s+')
        energy_line = re.compile(
            r"Final CI energies\s+=\s+(?P<energy>(?:[-+]?(?:\d+(?:\.\d*)?|\d*\.\d+)\s+)+)\s+"
        )
        conv_line = re.compile(r"(?:\s*root\s+\d+\s+\.+\s+(?:converged|unconverged)!\s*)+")
        sym_line = re.compile(
            r"\s*&+\s+KRCI calculation for symmetry no\.\s+(\d+)\s+&+\s+Number of CI roots for this symmetry\s+(\d+)\s+"
        )
        scf_line = re.compile(
            r"\* REACMO: Coefficients read from file DFCOEF - Total energy:\s+(?P<energy>[-+]?(\d+(\.\d*)?|\d*\.\d+))"
        )
        scf_line2 = re.compile(
            r"\s*Electronic energy\s*:\s*(?P<energy>[-+]?(\d+(\.\d*)?|\d*\.\d+))"
        )

        with open(self.filename) as fin:
            context = fin.read()

        energy_l = energy_line.findall(context, re.MULTILINE | re.DOTALL)
        conv_l = conv_line.findall(context, re.MULTILINE | re.DOTALL)
        sym_l = sym_line.findall(context, re.MULTILINE | re.DOTALL)

        energies = {}

        all_e = []
        for e_str in energy_l:
            e_roots = [float(e) for e in e_str.split()]
            all_e.append(e_roots)

        if len(sym_l) != len(all_e):
            warnings.warn("The shape of energy and symmetry are not " "the same! Check results!")
            return {}

        for i, sym_rt in enumerate(sym_l):
            tag_sym = sym_rt[0]
            nb_root = int(sym_rt[1])
            if nb_root != len(all_e[i]):
                warnings.warn(
                    "The number of root energies and the number of "
                    "roots are not the same! Check results!"
                )
                return {}

            for r in range(1, nb_root + 1):
                key = "_".join(["sym", str(tag_sym), "root", str(r)])
                energies[key] = float(all_e[i][r - 1])
                # energies['_'.join(['sym',str(tag_sym),'ave.'])] = sum(all_e[i])/len(all_e[i])

        convergeds = []
        for line in conv_l:
            converged = []
            for wrd in line.split("\n"):
                if not len(wrd.strip()):
                    continue
                if "unconverged" in wrd:
                    converged.append(False)
                elif "converged" in wrd:
                    converged.append(True)
                else:
                    print(wrd)
                    raise RuntimeError("There is no converged info")
            convergeds.append(converged)

        scf_energy = np.nan
        scf_l = scf_line.search(context)
        if scf_l:
            scf_energy = float(scf_l.group("energy"))
        else:
            scf_l2 = scf_line2.search(context)
            if scf_l2:
                scf_energy = float(scf_l2.group("energy"))
        self.energy_settings = {"scf_e": scf_energy, "ci_e": energies, "ci_converged": convergeds}

    def _parse_hartree_fock(self):
        """Deal with HF calculations

        Returns:
            None
        """

        pass

    def _parse_visual_module(self):
        """Deal with visual module calculations

        Returns:
            None
        """

        pass

    @property
    def calc_orbit(self):
        """Calculate the orbit.

        Returns
        -------
        dict or function
            If the calculation method is "CC", returns the CC orbit calculation function.
            If the calculation method is "CI", returns the CI orbit calculation function.
            Otherwise, returns a dictionary with "occ" and "vir" keys and values set to 0.

        """
        if self.inp.calc_method == "CC":
            return self._cc_calc_orbit
        elif self.inp.calc_method == "CI":
            return self.inp.ci_calc_orbit
        else:
            return {"occ": 0, "vir": 0}

    def as_dict(self) -> dict:
        """Return the object as a dictionary.

        Returns
        -------
        dict
            A dictionary representation of the object.

        """
        d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "energy_settings": self.energy_settings,
            "filename": self.filename,
            "task_type": self.task_type,
            "mol": self.mol.as_dict(),
            "inp": self.inp,
            "is_ok": self.is_ok,
        }
        return jsanitize(d, strict=True)



from enum import Enum
from math import inf

import numpy as np
from monty.json import MSONable, jsanitize

__all__ = ["OrbitalType", "AtomicOrbital", "MoleculeOrbitals"]


class OrbitalType(Enum):
    """
    An enumeration class representing the type of an orbital.

    Attributes
    ----------
    CLOSED_SHELL : int
        A closed shell orbital.
    OPEN_SHELL : int
        An open shell orbital.
    VIRTUAL_SHELL : int
        A virtual orbital.
    ALL_SHELL : int
        A combination of closed, open, and virtual orbitals.

    """

    CLOSED_SHELL = 0
    OPEN_SHELL = 1
    VIRTUAL_SHELL = 2
    ALL_SHELL = 3


class AtomicOrbital(MSONable):
    """
    A class representing an atomic orbital.

    Parameters
    ----------
    sym : str
        The symmetry of the orbital.
    type : OrbitalType
        The type of the orbital.
    energy : float
        The energy of the orbital in atomic units.
    degen : int
        The degeneracy of the orbital.
    frac : float
        The fractional occupancy of the orbital.

    Attributes
    ----------
    sym : str
        The symmetry of the orbital.
    type : OrbitalType
        The type of the orbital.
    energy : float
        The energy of the orbital in atomic units.
    degen : int
        The degeneracy of the orbital.
    frac : float
        The fractional occupancy of the orbital.

    Methods
    -------
    is_valid() -> bool
        Checks if the atomic orbital is valid.
    __str__() -> str
        Returns a string representation of the atomic orbital.
    __repr__() -> str
        Returns a string representation of the atomic orbital.
    __eq__(other: AtomicOrbital) -> bool
        Checks if two atomic orbitals are equal.
    as_dict() -> dict
        Returns a dictionary representation of the atomic orbital.
    from_dict(cls, d: dict) -> AtomicOrbital
        Returns an AtomicOrbital object created from a dictionary.

    """

    def __init__(self, sym: str, type: OrbitalType, energy: float, degen: int, frac: float):
        """
        Initialize an AtomicOrbital object.

        Parameters
        ----------
        sym : str
            The symmetry of the orbital.
        type : OrbitalType
            The type of the orbital.
        energy : float
            The energy of the orbital in atomic units.
        degen : int
            The degeneracy of the orbital.
        frac : float
            The fractional occupancy of the orbital.
        """
        self.sym = sym
        self.type = type
        self.energy = float(energy)
        self.frac = float(frac)
        self.degen = int(degen)

    def is_valid(self) -> bool:
        """
        Check if the atomic orbital is valid.

        Returns
        -------
        bool
            True if the atomic orbital is valid, False otherwise.
        """

        if self.type == OrbitalType.CLOSED_SHELL:
            return np.allclose(self.frac, 1.0)
        elif self.type is OrbitalType.VIRTUAL_SHELL:
            return np.allclose(self.frac, 0.0)
        else:
            return True

    def __str__(self) -> str:
        """
        Return a string representation of the atomic orbital.

        Returns
        -------
        str
            A string representation of the atomic orbital.
        """

        info_list = [self.sym, self.type, self.energy, self.degen, self.frac]
        return " ".join([str(info) for info in info_list])

    def __repr__(self) -> str:
        """
        Return a string representation of the atomic orbital.

        Returns
        -------
        str
            A string representation of the atomic orbital.
        """

        return self.__str__()

    def __eq__(self, other) -> bool:
        """
        Check if two atomic orbitals are equal.

        Parameters
        ----------
        other : AtomicOrbital
            The other atomic orbital.

        Returns
        -------
        bool
            True if the two atomic orbitals are equal, False
        """
        is_equal = True
        if self.sym != other.sym:
            is_equal = False
        if self.type != other.type:
            is_equal = False
        if abs(self.energy - other.energy) > 0.000001:
            is_equal = False
        if abs(self.frac - other.frac) > 0.01:
            is_equal = False
        if self.degen != other.degen:
            is_equal = False
        return is_equal

    def as_dict(self) -> dict:
        """
        Return a dictionary representation of the atomic orbital.

        Returns
        -------
        dict
            A dictionary with the following keys:
            - sym: str, the symmetry of the orbital
            - type: str, the type of the orbital
            - energy: float, the energy of the orbital in atomic units
            - degen: int, the degeneracy of the orbital
            - frac: float, the fractional occupancy of the orbital
        """

        ao_type = "closed_shell"
        if self.type is OrbitalType.OPEN_SHELL:
            ao_type = "open_shell"
        elif self.type is OrbitalType.VIRTUAL_SHELL:
            ao_type = "virtual_shell"

        d = {
            "degen": self.degen,
            "energy": self.energy,
            "frac": self.frac,
            "sym": self.sym,
            "type": ao_type,
        }
        return jsanitize(d, strict=True)

    @classmethod
    def from_dict(cls, d):
        """
        Create an `AtomicOrbital` object from a dictionary.

        Args:
            d (dict): The dictionary containing the `AtomicOrbital` object data.

        Returns:
            An `AtomicOrbital` object instance.
        """
        return AtomicOrbital(**d)


class MoleculeOrbitals(MSONable):
    """
    A class representing the orbitals of a molecule.

    Parameters
    ----------
    aos : list of AtomicOrbital objects, optional
        The atomic orbitals of the molecule.

    Attributes
    ----------
    aos : list of AtomicOrbital objects
        The atomic orbitals of the molecule.
    closed_shells : list of AtomicOrbital objects
        The closed-shell atomic orbitals of the molecule.
    open_shells : list of AtomicOrbital objects
        The open-shell atomic orbitals of the molecule.
    virtual_shells : list of AtomicOrbital objects
        The virtual atomic orbitals of the molecule.
    occ_open_shell : float
        The fractional occupancy of the open-shell orbitals.

    Methods
    -------
    analyse()
        Analyzes the atomic orbitals and separates them into different categories.
    type_found(test_type: OrbitalType, ao_type: OrbitalType) -> bool
        Checks if the atomic orbital type matches the specified type.
    nelec(e_min=-inf, e_max=inf, ao_type=OrbitalType.ALL_SHELL) -> int
        Counts the number of electrons in the specified energy range and atomic orbital type.
    nb_closed_elec(e_min=-inf, e_max=inf) -> int
        Counts the number of electrons in the closed-shell atomic orbitals within the specified energy range.
    nb_open_elec(e_min=-inf, e_max=inf) -> int
        Counts the number of electrons in the open-shell atomic orbitals within the specified energy range.
    nb_virtual_elec() -> int
        Counts the number of electrons in the virtual atomic orbitals.
    nao(e_min=-inf, e_max=inf, ao_type=OrbitalType.ALL_SHELL) -> int
        Counts the number of atomic orbitals in the specified energy range and atomic orbital type.
    nb_open_ao(e_min=-inf, e_max=inf) -> int
        Counts the number of open-shell atomic orbitals within the specified energy range.
    nb_virtual_ao(e_min=-inf, e_max=inf) -> int
        Counts the number of virtual atomic orbitals within the specified energy range.
    nb_closed_ao(e_min=-inf, e_max=inf) -> int
        Counts the number of closed-shell atomic orbitals within the specified energy range.
    get_ao_and_elec(e_min, e_max)
        Returns the number of electrons and atomic orbitals in different categories within the specified energy range.
    as_dict() -> dict
        Returns a dictionary representation of the MoleculeOrbitals object.
    """

    def __init__(self, aos=None):
        """
        Initialize a MoleculeOrbitals object.

        Parameters
        ----------
        aos : list of AtomicOrbital objects, optional
            The atomic orbitals of the molecule.
        """
        self.aos = aos
        self.closed_shells = []
        self.open_shells = []
        self.virtual_shells = []
        self.occ_open_shell = 0.0

        if self.aos:
            self.analyse()

    def analyse(self):
        """
        Analyze the atomic orbitals and separates them into different categories.
        """
        # TODO: can we get if this is an atomic calculation
        for ao in self.aos:
            # TODO: do we need to check degeneration?
            if ao.type is OrbitalType.CLOSED_SHELL:
                self.closed_shells.append(ao)
            elif ao.type is OrbitalType.OPEN_SHELL:
                self.open_shells.append(ao)
            elif ao.type is OrbitalType.VIRTUAL_SHELL:
                self.virtual_shells.append(ao)
            else:
                raise RuntimeError("Unknown orbital type !")

    def type_found(self, test_type, ao_type=OrbitalType.ALL_SHELL):
        """
        Check if the atomic orbital type matches the specified type.

        Parameters
        ----------
        test_type : OrbitalType
            The atomic orbital type to check against.
        ao_type : OrbitalType, optional
            The specified atomic orbital type to match against. Defaults to OrbitalType.ALL_SHELL.

        Returns
        -------
        bool
            True if the atomic orbital type matches the specified type, False otherwise.
        """
        if ao_type is OrbitalType.ALL_SHELL:
            return True
        else:
            return test_type is ao_type

    def nelec(self, e_min=-inf, e_max=inf, ao_type=OrbitalType.ALL_SHELL) -> int:
        """
        Count the number of electrons in the specified energy range and atomic orbital type.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.
        ao_type : OrbitalType, optional
            The specified atomic orbital type to count. Defaults to OrbitalType.ALL_SHELL.

        Returns
        -------
        int
            The number of electrons located in the energy range.
        """
        nb_elec = sum(
            [
                ao.degen * ao.frac
                for ao in self.aos
                if self.type_found(ao.type, ao_type) and e_min <= ao.energy <= e_max
            ]
        )
        return round(nb_elec)

    def nb_closed_elec(self, e_min=-inf, e_max=inf) -> int:
        """
        Count the number of electrons in the closed-shell atomic orbitals within the specified energy range.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.

        Returns
        -------
        int
            The number of electrons located in the energy range in the closed-shell atomic orbitals.
        """
        return self.nelec(e_min, e_max, ao_type=OrbitalType.CLOSED_SHELL)

    def nb_open_elec(self, e_min=-inf, e_max=inf) -> int:
        """
        Count the number of electrons in the open-shell atomic orbitals within the specified energy range.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.

        Returns
        -------
        int
            The number of electrons located in the energy range in the open-shell atomic orbitals.
        """
        return self.nelec(e_min, e_max, ao_type=OrbitalType.OPEN_SHELL)

    def nb_virtual_elec(self) -> int:
        """
        Count the number of electrons in the virtual atomic orbitals.

        Returns
        -------
        int
            The number of electrons in the virtual atomic orbitals.
        """
        return 0

    def nao(self, e_min=-inf, e_max=inf, ao_type=OrbitalType.ALL_SHELL) -> int:
        """
        Count the number of atomic orbitals in the specified energy range and atomic orbital type.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.
        ao_type : OrbitalType, optional
            The specified atomic orbital type to count. Defaults to OrbitalType.ALL_SHELL.

        Returns
        -------
        int
            The number of atomic orbitals located in the energy range.
        """
        nb_ao = sum(
            [
                ao.degen
                for ao in self.aos
                if self.type_found(ao.type, ao_type) and e_min <= ao.energy <= e_max
            ]
        )
        return round(nb_ao)

    def nb_open_ao(self, e_min=-inf, e_max=inf) -> int:
        """
        Count the number of open-shell atomic orbitals within the specified energy range.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.

        Returns
        -------
        int
            The number of open-shell atomic orbitals located in the energy range.
        """
        return self.nao(e_min, e_max, ao_type=OrbitalType.OPEN_SHELL)

    def nb_virtual_ao(self, e_min=-inf, e_max=inf) -> int:
        """
        Count the number of virtual atomic orbitals within the specified energy range.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.

        Returns
        -------
        int
            The number of virtual atomic orbitals located in the energy range.
        """
        return self.nao(e_min, e_max, ao_type=OrbitalType.VIRTUAL_SHELL)

    def nb_closed_ao(self, e_min=-inf, e_max=inf) -> int:
        """
        Count the number of closed-shell atomic orbitals within the specified energy range.

        Parameters
        ----------
        e_min : float, optional
            The minimum energy of the range. Defaults to -inf.
        e_max : float, optional
            The maximum energy of the range. Defaults to inf.

        Returns
        -------
        int
            The number of closed-shell atomic orbitals located in the energy range.
        """
        return self.nao(e_min, e_max, ao_type=OrbitalType.CLOSED_SHELL)

    def get_ao_and_elec(self, e_min, e_max):
        """
        Obtain the number of electrons and atomic orbitals in the specified energy range.

        Parameters
        ----------
        e_min : float
            The minimum energy of the range.
        e_max : float
            The maximum energy of the range.

        Returns
        -------
        tuple
            A tuple containing the number of closed-shell electrons, open-shell electrons, total electrons,
            closed-shell atomic orbitals, open-shell atomic orbitals, and virtual atomic orbitals.
        """
        # obtain the number of electrons
        nb_closed_elec = self.nb_closed_elec(e_min, e_max)
        nb_open_elec = self.nb_open_elec(e_min, e_max)
        nb_total_elec = self.nelec(e_min, e_max)

        # obtain orbit info
        ## res_orbits = [nb_closed_shells, nb_open_shells, nb_virtual_shells]
        nb_closed_shell = self.nb_closed_ao(e_min, e_max)
        nb_open_shell = self.nb_open_ao(e_min, e_max)
        nb_vir_shell = self.nb_virtual_ao(e_min, e_max)
        return (
            nb_closed_elec,
            nb_open_elec,
            nb_total_elec,
            nb_closed_shell,
            nb_open_shell,
            nb_vir_shell,
        )

    def as_dict(self) -> dict:
        """
        Return a dictionary representation of the MoleculeOrbitals object.

        Returns
        -------
        dict
            A dictionary containing the closed-shell, open-shell, and virtual atomic orbitals, as well as
            the occupation number of open-shell atomic orbitals.
        """
        d = {
            "closed_shells": self.closed_shells,
            "occ_open_shell": self.occ_open_shell,
            "open_shells": self.open_shells,
            "virtual_shells": self.virtual_shells,
        }
        return jsanitize(d, strict=True)



import numpy as np
from monty.json import MSONable, jsanitize

__all__ = ["Element"]
# fmt: off
_PERIODIC_TABLE = [
    'dummy', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',
    'Cl', 'Ar','K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge',
    'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
    'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd',
    'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
    'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm',
    'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn',
    'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og'
]
# fmt: on

_SPECIAL_ELEMENTS = {
    24: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^1 3d^5",
    29: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^1 3d^10",
    41: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^4",
    42: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^5",
    44: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^7",
    45: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^8",
    46: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 4s^10",
    47: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^10",
    57: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^10 5p^6 6s^2 5d^1",
    58: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^10 5p^6 6s^2 4f^1 5d^1",
    64: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^1 4d^10 5p^6 6s^2 4f^7 5d^1",
    78: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^1 4f^14 5d^9",
    79: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^1 4f^14 5d^10",
    89: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 6d^1",
    90: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 6d^2",
    91: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 5f^2 6d^1",
    92: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 5f^3 6d^1",
    93: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 5f^4 6d^1",
    96: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 5f^7 6d^1",
    103: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^2 5f^14 7p^1",
    110: "1s^2 2s^2 2p^6 3s^2 3p^6 4s^2 3d^10 4p^6 5s^2 4d^10 5p^6 6s^2 4f^14 5d^10 6p^6 7s^1 5f^14 6d^9",
}


class Element(MSONable):
    """
    Represents a chemical element in the periodic table.

    Attributes
    ----------
    atomic_number : int
        The atomic number of the element.
    symbol : str
        The symbol of the element.
    """

    ROW_SIZES = (2, 8, 8, 18, 18, 32, 32)

    def __init__(self, element):
        """
        Initialize the Element object.

        Parameters
        ----------
        element : str or int
            The symbol or atomic number of the element.
        """
        if isinstance(element, (str, int, np.integer)):
            idx, symbol = self._get_element(element)
        else:
            raise ValueError(f"Expected a <str> or <int>, got: {type(element):s}")

        self.atomic_number = idx
        self.symbol = symbol

    @staticmethod
    def _get_element(element):
        """
        Get the element by the given id (either symbol or atomic number).

        Parameters
        ----------
        element : str or int
            The symbol or atomic number of the element.

        Returns
        -------
        tuple
            A tuple containing the atomic number and symbol of the element.
        """

        def get_symbol(Z):
            """
            Get element by given index.
            """
            if isinstance(Z, (int, np.integer)):
                if Z < 0 or Z > 118:
                    raise ValueError(f"Wrong atomic number: {Z}")
                return _PERIODIC_TABLE[Z]
            else:
                raise RuntimeError("Type id should be int")

        def get_id(symbol):
            """
            Get element number by given element type.
            """

            if symbol not in _PERIODIC_TABLE:
                raise ValueError(f"Wrong atomic symbol {symbol}")
            idx = _PERIODIC_TABLE.index(symbol)
            if idx >= 0:
                return idx
            else:
                raise RuntimeError("Wrong element type!")

        if isinstance(element, str):
            return get_id(element), element
        else:
            return element, get_symbol(element)

    @property
    def block(self):
        """
        Get the block character (s, p, d, or f) of the element.

        Returns
        -------
        str
            The block character of the element.
        """
        if (self.is_actinoid or self.is_lanthanoid) and self.atomic_number not in [
            71,
            103,
        ]:
            return "f"
        if self.is_actinoid or self.is_lanthanoid:
            return "d"
        if self.group in [1, 2]:
            return "s"
        if self.group in range(13, 19):
            return "p"
        if self.group in range(3, 13):
            return "d"
        raise ValueError("unable to determine block")

    @property
    def is_lanthanoid(self):
        """
        Check if the element is a lanthanoid.

        Returns
        -------
        bool
            True if the element is a lanthanoid, False otherwise.
        """
        return 56 < self.atomic_number < 72

    @property
    def is_actinoid(self):
        """
        Check if the element is an actinoid.

        Returns
        -------
        bool
            True if the element is an actinoid, False otherwise.
        """
        return 88 < self.atomic_number < 104

    @property
    def row(self):
        """
        Get the periodic table row of the element.

        Returns
        -------
        int
            The row of the element in the periodic table.
        """
        z = self.Z
        total = 0
        if 57 <= z <= 71:
            return 8
        if 89 <= z <= 103:
            return 9
        for i, size in enumerate(Element.ROW_SIZES):
            total += size
            if total >= z:
                return i + 1
        return 8

    @property
    def group(self):
        """
        Get the periodic table group of the element.

        Returns
        -------
        int
            The group of the element in the periodic table.
        """
        z = self.atomic_number
        if (self.is_actinoid or self.is_lanthanoid) and z not in [71, 103]:
            return None

        if z == 1:
            return 1
        if z == 2:
            return 18
        if 3 <= z <= 18:
            if (z - 2) % 8 == 0:
                return 18
            if (z - 2) % 8 <= 2:
                return (z - 2) % 8
            return 10 + (z - 2) % 8

        if 19 <= z <= 54:
            if (z - 18) % 18 == 0:
                return 18
            return (z - 18) % 18

        assert z >= 54
        rest = (z - 54) % 32
        # group 18
        if rest == 0:
            return 18
        # group 1 and 2
        elif 1 <= rest <= 2:
            return rest
        #  group 3-12
        elif 2 < rest < 16:
            return None
        # group 13-17
        elif rest >= 16:
            return rest - 14
        else:
            raise ValueError(f"You found a bug!")

    @property
    def group_symbol(self):
        """
        Get the group symbol of the element.

        Returns
        -------
        str
            The group symbol of the element.
        """
        return f"group-{self.group}"

    @property
    def Z(self):
        """
        Get the atomic number of the element.

        Returns
        -------
        int
            The atomic number of the element.
        """
        return self.atomic_number

    @property
    def period(self):
        """
        Get the period of the element in the periodic table.

        Returns
        -------
        int
            The period of the element in the periodic table.
        """
        return self.row

    def as_dict(self) -> dict:
        """
        Get a dictionary representation of the Element object.

        Returns
        -------
        dict
            The dictionary representation of the Element object.
        """
        d = {
            "Z": self.Z,
            "block": self.block,
            "row": self.row,
            "group": self.group,
            "group_symbol": self.group_symbol,
            "period": self.period,
        }
        return jsanitize(d)

    def get_elec_config(self):
        """
        Get the electron configuration of the element.

        Returns
        -------
        str
            The electron configuration of the element.
        """
        n = self.Z
        if n in _SPECIAL_ELEMENTS.keys():
            return _SPECIAL_ELEMENTS[n]

        rule = "1s 2s 2p 3s 3p 4s 3d 4p 5s 4d 5p 6s 4f 5d 6p 7s 5f 6d 7p"
        nb_dict = {"s": 2, "p": 6, "d": 10, "f": 14}
        orbitals = [(i, nb_dict[i[-1]]) for i in rule.split()]
        output = []

        for orbital, size in orbitals:
            k = min(size, n)
            output.append("%s^%d" % (orbital, k))
            n -= k
            if n <= 0:
                break

        orbital_info = " ".join(output)
        return orbital_info

    def get_dirac_shell_str(self, return_dict=False):
        """
        Returns the electron configurations in DIRAC according to the number of electrons.

        Parameters
        ----------
        return_dict : bool, optional
            Whether to return the results as a dictionary or string. Default is False.

        Returns
        -------
        str or dict
            The electron configurations in DIRAC.

        Notes
        -----
        For example:

        - for NSYM = 1
            ```text
            .CLOSED SHELL
            2
            .OPEN SHELL
            1
            1/2
            ```

        - or for NSYM != 1
            ```text
            .CLSED SHELL
            2 0
            .OPEN SHELL
            1
            1/2,0
            ```
        """

        def get_shell(elec_config, NSYM=1):
            """
            Determines the number of electrons in each shell and subshell for a given electronic configuration.

            Parameters
            ----------
            elec_config : str
                The electronic configuration of the atom. For example, "1s2 2s2 2p6" represents the configuration of Neon.
            NSYM : int, optional
                The number of symmetry operations, defaults to 1.

            Returns
            -------
            cs_str : str or None
                The string representation of the number of electrons in closed shells, or None if there are no closed shells.
            os_str : str or None
                The string representation of the number of electrons in open shells, or None if there are no open shells.

            Notes
            -----
            This function uses the Aufbau principle to determine the electron configuration. The order of orbitals is as follows:
            "1s 2s 2p 3s 3p 4s 3d 4p 5s 4d 5p 6s 4f 5d 6p 7s 5f 6d 7p"
            The maximum number of electrons that can be in each subshell is as follows:
            s: 2
            p: 6
            d: 10
            f: 14
            Any subshell with less than the maximum number of electrons is considered open shell, otherwise it is considered closed shell.

            Examples
            --------
            >>> get_shell("1s2 2s2 2p6")
            ('10', None)

            >>> get_shell("1s2 2s2 2p5")
            (None, '1/2,0\\n1/2,0\\n2/6,0')

            """
            rule = "1s 2s 2p 3s 3p 4s 3d 4p 5s 4d 5p 6s 4f 5d 6p 7s 5f 6d 7p"
            nb_dict = {"s": 2, "p": 6, "d": 10, "f": 14}
            cs_even, cs_odd = 0, 0
            os_even, os_odd = [], []
            for i in elec_config.split():
                # 1s^2
                t, nb = i.split("^")
                if int(nb) < nb_dict[t[-1]]:
                    # open_shell
                    if t[-1] in ["s", "d"]:
                        # even
                        os_even.append((t, int(nb)))
                    else:
                        os_odd.append((t, int(nb)))
                else:
                    if t[-1] in ["s", "d"]:
                        # even
                        cs_even += int(nb)
                    else:
                        cs_odd += int(nb)

            cs_total = cs_even + cs_odd
            os_total = os_even + os_odd
            os_total = sorted(os_total, key=lambda x: rule.index(x[0]))
            nb_os = len(os_total)

            cs_str, os_str = [], []
            if NSYM == 1:
                if cs_total == 0:
                    cs_str = None
                else:
                    cs_str.append(f"{cs_total}")
                if not len(os_total):
                    os_str = None
                else:
                    os_str.append(f"{nb_os}")
                    for i in os_total:
                        os_str.append(f"{i[-1]}/{nb_dict[i[0][1]]}")
            else:
                # for closed shell: 10, 10
                # for open shell:
                # 2
                # 5/10,0
                # 1/2,0
                if cs_total:
                    cs_str.append(f"{cs_even} {cs_odd}\n")
                else:
                    cs_str = None

                if nb_os:
                    os_str.append(f"{nb_os}\n")
                    for i in os_total:
                        if i[0][-1] in ["s", "d"]:
                            os_str.append(f"{i[-1]}/{nb_dict[i[0][1]]},{0}\n")
                        else:
                            os_str.append(f"{i[-1]}/{0},{nb_dict[i[0][1]]}\n")
                else:
                    os_str = None
            if cs_str:
                cs_str = "\n".join(cs_str)
            if os_str:
                os_str = "\n".join(os_str)
            return cs_str, os_str

        elec_config = self.get_elec_config()
        closed_shell, open_shell = get_shell(elec_config)

        res_str = []
        res_dict = {}
        res_str.append(f"# {self.symbol} atom " + f"and Econf: {elec_config}")
        if closed_shell:
            res_str.append(".CLOSED SHELL")
            res_str.append(closed_shell)
            res_dict["CLOSED SHELL"] = closed_shell
        if open_shell:
            res_str.append(".OPEN SHELL")
            res_str.append(open_shell)
            res_dict["OPEN SHELL"] = open_shell

        res_str = "\n".join(res_str)
        if return_dict:
            return res_dict
        else:
            return res_str



import glob
import os
import warnings

import numpy as np
from scipy.linalg import lstsq
from scipy.special import factorial
from monty.os import cd

from pydirac.analysis.utility import get_keyword, get_orbital_info
from pydirac.io.outputs import Output

__all__ = [
    "FiniteFieldLstsqSolver",
    "PolarizabilityCalculator",
    "check_results",
    "get_polarizability",
    "get_polarizability_from_output_list",
    "do_one_basis",
    "is_valid",
]


def check_results(x_svd, errors, nb_coeffs):
    """
    Check the results of the least squares fitting for validity.

    Parameters
    ----------
    x_svd : ndarray
        Solution to the least squares problem.
    errors : ndarray
        Standard errors of the solution.
    nb_coeffs : int
        The number of fitted coefficients required.

    Returns
    -------
    bool
        True if the results are valid, False otherwise.
    """
    if nb_coeffs > 1:
        if np.isclose(x_svd[1], 0.0):
            return np.allclose(x_svd[1:], 0.0) and np.allclose(errors[1:], 0.0)
        if x_svd[1] < 0 or errors[1] / x_svd[1] < 1:
            return False
    return True


class FiniteFieldLstsqSolver:
    """
    A class to solve the least squares problem for polarizability calculations using finite field method.
    """

    def __init__(
        self, energy, field, nb_coeffs, rcond, threshold, calc_type, check_hyper_polar
    ):
        """
        Initialize the FiniteFieldLstsqSolver instance.

        Parameters
        ----------
        energy : array_like
            Array of energy values corresponding to different field strengths.
        field : array_like
            Array of field strengths.
        nb_coeffs : int
            Number of coefficients to use for the least squares fitting. Must be between 1 and 3.
        rcond : float
            Cut-off ratio for small singular values in the least squares solution.
        threshold : float
            Threshold value for field strength to include in the fitting.
        calc_type : {'dipole', 'quadrupole'}
            Type of polarizability calculation to perform.
        check_hyper_polar : bool
            Whether to check for hyperpolarizability during fitting.
        """
        assert calc_type in ["dipole", "quadrupole"]
        self.energy = np.asarray(energy)
        self.field = np.asarray(field)
        self.nb_coeffs = nb_coeffs
        self.rcond = rcond
        self.threshold = threshold
        self.calc_type = calc_type
        self.check_hyper_polar = check_hyper_polar

    def calc_dipole_mat(self, fields):
        """
        Calculate the matrix for dipole polarizability fitting.

        Parameters
        ----------
        fields : array_like
            Array of field strengths.

        Returns
        -------
        ndarray
            The matrix for dipole polarizability fitting.
        """
        f_list = [
            -1 / factorial(2 * i + 2) * fields ** (2 * i + 2)
            for i in range(self.nb_coeffs)
        ]
        return np.column_stack(f_list)

    def calc_quadrupole_mat(self, fields):
        """
        Calculate the matrix for quadrupole polarizability fitting.

        Parameters
        ----------
        fields : array_like
            Array of field strengths.

        Returns
        -------
        ndarray
            The matrix for quadrupole polarizability fitting.
        """
        if self.nb_coeffs == 1:
            f_list = [1 / 2 * fields]
        elif self.nb_coeffs == 2:
            f_list = [1 / 2 * fields, -1 / 8 * fields**2]
        elif self.nb_coeffs == 3:
            f_list = [1 / 2 * fields, -1 / 8 * fields**2, 1 / 24 * fields**3]
        else:
            raise RuntimeError(
                f"nb_coeffs={self.nb_coeffs} should be less than 4 and larger than 0"
            )
        return np.column_stack(f_list)

    def calc_mat(self, fields):
        """
        Calculate the fitting matrix based on the calculation type.

        Parameters
        ----------
        fields : array_like
            Array of field strengths.

        Returns
        -------
        ndarray
            The fitting matrix.
        """
        if self.calc_type == "dipole":
            return self.calc_dipole_mat(fields)
        elif self.calc_type == "quadrupole":
            return self.calc_quadrupole_mat(fields)
        else:
            raise NotImplementedError()

    def preprocess_data(self):
        """
        Preprocess the data by filtering and sorting the field and energy arrays.

        Returns
        -------
        sorted_field : ndarray
            Sorted array of filtered field strengths.
        sorted_energy : ndarray
            Sorted array of filtered energy values.
        e_ref : float
            Reference energy value at zero field.

        Raises
        ------
        ValueError
            If no field value is close to zero.
        """
        zero_field_mask = np.isclose(self.field, 0.0)
        if np.any(zero_field_mask):
            e_ref = self.energy[zero_field_mask][0]
        else:
            raise ValueError("No field value is close to zero.")
        mask = (self.field != 0.0) & (self.field <= self.threshold + 1e-8)
        reduced_field = self.field[mask]
        reduced_energy = self.energy[mask]
        sort_indices = np.argsort(reduced_field)
        sorted_field = reduced_field[sort_indices]
        sorted_energy = reduced_energy[sort_indices]
        return sorted_field, sorted_energy, e_ref

    def solve_least_squares(self, A, b):
        """
        Solve the least squares problem and return the solution, residual, and normalization factors.

        Parameters
        ----------
        A : ndarray
            The fitting matrix.
        b : ndarray
            The energy difference vector.

        Returns
        -------
        x_svd : ndarray
            Solution to the least squares problem.
        residual : float
            The residual sum of squares.
        norms : ndarray
            Normalization factors for the columns of the fitting matrix.
        """
        norms = np.linalg.norm(A, axis=0)
        A_normalized = A / norms
        x_svd_normalized, residuals, _, _ = lstsq(A_normalized, b, cond=self.rcond)

        if A.shape[0] > A.shape[1]:
            assert np.ndim(b) == 1
            assert np.isscalar(residuals)
            residual = residuals
        else:
            assert A.shape[1] == b.shape[0]
            # an empty array is returned
            assert len(residuals) == 0
            residual = 0.0
        x_svd = x_svd_normalized / norms

        ## Old implementation
        # self.rcond = None
        # norms = 1
        # A_normalized = A / norms
        # x_svd_normalized, residuals, _, _ = np.linalg.lstsq(
        #     A_normalized, b, rcond=self.rcond
        # )
        # if len(residuals) == 1:
        #     residual = residuals[0]
        # elif len(residuals) == 0:
        #     residual = 0
        # x_svd = x_svd_normalized / norms
        return x_svd, residual, norms

    def calculate_covariance_matrix(self, A, residual, degrees_of_freedom):
        """
        Calculate the covariance matrix for the least squares solution.

        Parameters
        ----------
        A : ndarray
            The fitting matrix.
        residual : float
            The residual sum of squares.
        degrees_of_freedom : int
            Degrees of freedom for the fitting.

        Returns
        -------
        ndarray
            The covariance matrix.
        """
        if degrees_of_freedom == 0:
            cov_matrix = np.zeros((A.shape[1], A.shape[1]))
        else:
            MSE = residual / degrees_of_freedom
            cov_matrix = np.linalg.inv(np.dot(A.T, A)) * MSE
        return cov_matrix

    def check_results(self, x_svd, errors):
        """
        Check the results of the least squares fitting for validity.

        Parameters
        ----------
        x_svd : ndarray
            Solution to the least squares problem.
        errors : ndarray
            Standard errors of the solution.

        Returns
        -------
        bool
            True if the results are valid, False otherwise.
        """
        return check_results(x_svd, errors, self.nb_coeffs)

    def fit(self):
        """
        Perform the fitting process and return the fitted coefficients and their standard errors.

        Returns
        -------
        x_svd : ndarray
            Fitted coefficients.
        standard_errors : ndarray
            Standard errors of the fitted coefficients.

        Raises
        ------
        RuntimeError
            If an unexpected condition occurs during fitting.
        """
        sorted_field, sorted_energy, e_ref = self.preprocess_data()
        A = self.calc_mat(sorted_field)
        b = sorted_energy - e_ref
        x_svd, residual, _ = self.solve_least_squares(A, b)
        degrees_of_freedom = len(sorted_energy) - len(x_svd)
        cov_matrix = self.calculate_covariance_matrix(A, residual, degrees_of_freedom)
        standard_errors = np.sqrt(np.diag(cov_matrix))

        if self.check_hyper_polar:
            if self.check_results(x_svd, standard_errors):
                return x_svd, standard_errors
            elif self.nb_coeffs > 1:
                x_svd, standard_errors = FiniteFieldLstsqSolver(
                    self.energy,
                    self.field,
                    1,
                    self.rcond,
                    self.threshold,
                    self.calc_type,
                    self.check_hyper_polar,
                ).fit()
                tmp_svd = np.zeros((self.nb_coeffs,))
                tmp_errors = np.zeros((self.nb_coeffs,))
                tmp_svd[0] = x_svd[0]
                tmp_errors[0] = standard_errors[0]
                return tmp_svd, tmp_errors
            else:
                raise RuntimeError("You found a bug!")
        else:
            return x_svd, standard_errors


class PolarizabilityCalculator:
    """
    A class to calculate dipole polarizability and quadrupole polarizability using finite difference
    method.

    Parameters
    ----------
    calc_type : str, optional
        The type of calculation to perform. Must be one of "dipole" or "quadrupole".
        Default is "dipole".
    nb_coeffs : int, optional
        The number of coefficients to use for the calculation. Must be 2 or 3.
        Default is 2.

    Attributes
    ----------
    calc_type : str
        The type of calculation to perform. Must be one of "dipole" or "quadrupole".
    nb_coeffs : int
        The number of coefficients to use for the calculation.
    """

    def __init__(self, calc_type="dipole", nb_coeffs=2):
        """
        Initialize the PolarizabilityCalculator instance.

        Parameters
        ----------
        calc_type : str, optional
            The type of calculation to perform. Must be one of "dipole" or "quadrupole".
            Default is "dipole".
        nb_coeffs : int, optional
            The number of coefficients to use for the calculation. Must be 2 or 3.
            Default is 2.
        """
        _ct = calc_type.lower()
        if _ct in ["d", "dipole"]:
            _ct = "dipole"
        elif _ct in ["q", "quad", "quadrupole"]:
            _ct = "quadrupole"
        else:
            raise ValueError("Wrong calculation type!")
        self.calc_type = _ct

        if nb_coeffs not in (1, 2, 3):
            raise ValueError("nb_coeffs must be 1, 2 or 3!")
        self.nb_coeffs = nb_coeffs

    def get_svd_from_array(
        self, energy, field, rcond=1e-10, threshold=None, check_hyper_polar=False
    ):
        return FiniteFieldLstsqSolver(
            energy,
            field,
            self.nb_coeffs,
            rcond,
            threshold,
            self.calc_type,
            check_hyper_polar,
        ).fit()


def get_polarizability(
    dirname: str = "./",
    calc_dir_patters=None,
    deepth: int = 0,
    verbos=False,
    threshold=None,
) -> dict:
    """Get polarizability from a directory

    A calculation directory may contain some information as followings:
        (1) no more calculation directory (calc_dir_list) but several output files (curr_dir_output_list)
        (2) several output files (curr_dir_output_list) with other calculation directories (calc_dir_list)
        (3) several output files (curr_dir_output_list) with subdirectories (sub_dir_list) --> call self again when deepth > 0
        (4) only calculations directories (calc_dir_list)
        (5) calculations directories (calc_dir_list) and (sub_dir_list)
        (6) only subdirectories (sub_dir_list)
        (7) all (calc_dir_list), (curr_dir_output_list) and (sub_dir_list)

    Args:
        dirname: dirname
        suffix:

    Returns:
        None
    """
    # deepth = 0: only calc current output files
    if calc_dir_patters is None:
        calc_dir_patters = ["*dyall*"]

    if isinstance(calc_dir_patters, (str, list)):
        if isinstance(calc_dir_patters, str):
            calc_dir_patters = [calc_dir_patters]
    else:
        raise TypeError("calc_dir_patters can only be <str> or <list>")

    deepth = 0 if deepth < 0 else deepth
    do_sub_dir = deepth > 0

    # for debug
    if verbos:
        print(f"cd {dirname}")

    all_res = {}
    curr_dir_output_lis = []
    calc_dir_output_lis = []
    with cd(dirname):
        # Step 1. deal with all current output files
        # ------------------------------
        for f in glob.glob("*.out"):
            obj = Output(filename=f)
            if obj.is_ok:
                curr_dir_output_lis.append(obj)

        # Step 2. deal with all calc_dir
        # ------------------------------
        clc_dirs = []
        for ptn in calc_dir_patters:
            clc_dirs.extend([d for d in glob.glob("*" + ptn + "*") if os.path.isdir(d)])
        # if has calc_dir then I would try
        for clc_d in clc_dirs:
            with cd(clc_d):
                outs = glob.glob("*.out")
                if len(outs) > 1:
                    warnings.warn(
                        "there are more than two output file in this directory {}, and we take "
                        "the first one {}".format(clc_d, outs[0])
                    )
                elif len(outs) == 1:
                    obj = Output(filename=outs[0])
                    if obj.is_ok:
                        calc_dir_output_lis.append(obj)
                else:
                    warnings.warn(f"there is no output file in this directory {clc_d}")

        all_res["curr_dir"] = {}
        if is_valid(curr_dir_output_lis):
            e_dict = get_polarizability_from_output_list(
                dirname,
                curr_dir_output_lis,
                tag="curr_dir",
                verbos=verbos,
                threshold=threshold,
            )
            all_res["curr_dir"] = e_dict

        all_res["calc_dir"] = {}
        if is_valid(calc_dir_output_lis):
            e_dict = get_polarizability_from_output_list(
                dirname,
                calc_dir_output_lis,
                tag="clc_dir",
                verbos=verbos,
                threshold=threshold,
            )
            all_res["calc_dir"] = e_dict

        # Step 3. deal with all sub_dir
        # ------------------------------
        all_res["sub_dir"] = {}
        if do_sub_dir:
            sub_dirs = [
                d for d in glob.glob("*") if os.path.isdir(d) and d not in clc_dirs
            ]

            for sd in sub_dirs:
                res = get_polarizability(
                    os.path.join(dirname, sd),
                    calc_dir_patters,
                    deepth - 1,
                    verbos=verbos,
                    threshold=threshold,
                )
                if sd not in all_res["sub_dir"] and len(res) > 0:
                    all_res["sub_dir"][os.path.basename(sd)] = res
    return all_res


def do_one_basis(output_lis, threshold=None):
    if not len(output_lis):
        warnings.warn("there is no valid output file here")
        return

    # check the calc_type from the first output file
    ct = output_lis[0].inp.calc_type
    calc_orbit = output_lis[0].calc_orbit
    pc = PolarizabilityCalculator(calc_type=ct)

    # maybe we just use a dict to restore all information
    energies = {}
    # cause for different type calculations, the electric fields are the same
    fields = []

    for o in output_lis:
        if o.inp.calc_type != ct:
            warnings.warn(
                "we found a error output whose calc_type {} is "
                "different with the first one {}".format(o.inp.calc_type, ct)
            )
            continue

        # CC or CI
        fields.append(float(o.inp.electric_field))
        if o.inp.calc_method == "CC":
            for k, v in o.energy_settings.items():
                if k not in energies.keys():
                    energies[k] = []
                energies[k].append(v)

        elif o.inp.calc_method == "CI":
            # check all roots converged
            for i in o.energy_settings["ci_converged"]:
                if not np.all(np.asarray(i)):
                    raise RuntimeError("Not all roots are converged!")

            for k, v in o.energy_settings["ci_e"].items():
                if k not in energies.keys():
                    energies[k] = []
                energies[k].append(v)

            if "scf_e" not in energies:
                energies["scf_e"] = []
            energies["scf_e"].append(o.energy_settings["scf_e"])
        else:
            # raise NotImplementedError
            continue

    # remove items where the length of energy is not equal to the one of field
    del_k_lis = []
    for k, v in energies.items():
        if len(v) != len(fields):
            warnings.warn(
                'The length of energy set "{}" {} does '
                "not equal the length of field {}".format(k, len(v), len(fields))
            )
            del_k_lis.append(k)
    for k in del_k_lis:
        energies.pop(k)

    # {'scf': alpha_scf, 'mp2':alpha_scf, 'ccsd':alpha_ccsd}
    res = {}
    res_error = {}
    for k, energy in energies.items():
        polar_key = k.strip("_e")
        _polar, _error = pc.get_svd_from_array(energy, fields, threshold=threshold)
        res[polar_key] = _polar
        res_error[polar_key] = _error
    res_all = {
        "energy": {"energies": energies, "fields": fields},
        "polar": res,
        "polar_error": res_error,
    }
    return res_all


def get_polarizability_from_output_list(
    dirname, output_lis, tag=None, verbos=True, threshold=None
):
    all_basis_res = {}
    for o in output_lis:
        task_type, orbit = o.task_type, o.calc_orbit
        obt_info = (
            get_orbital_info(orbit["occ"], orbit["vir"]) if len(orbit) else "null"
        )
        k = get_keyword(o.mol.molecule.atomic_info.symbol, task_type, obt_info)
        if k not in all_basis_res:
            all_basis_res[k] = []
        all_basis_res[k].append(o)

    e_res = {}
    for k, v in all_basis_res.items():
        one_res = do_one_basis(v, threshold=threshold)
        if one_res:
            e_res[k] = one_res
    return e_res


def is_valid(output_lis, verbos=False):
    """Check whether a output_lis is valid

    Args:
        output_lis: a list of Output obj

    Returns:
        True or False
    """

    if len(output_lis) < 3:
        if verbos:
            warnings.warn("the nubmer of output objects is less than 3")
        return False

    # check all there file if they are all 'CC' or 'CI' calculations
    task_record = {}
    for o in output_lis:
        if o.inp.calc_method in ["CC", "CI"]:
            if not o.task_type in task_record:
                task_record[o.task_type] = 1
            else:
                task_record[o.task_type] += 1
    for v in task_record.values():
        if v >= 3:
            return True
    else:
        if verbos:
            print(task_record)
            warnings.warn(
                "the maximum of output objects with the same type is less than 3"
            )
        return False



import argparse
import sys

from pydirac.cli.pypam_atom_db import get_atomDB
from pydirac.core.basis import get_custom_basis_from_ele


def get_inp():
    pass


def main():
    pypam_parser = argparse.ArgumentParser(
        description="pypam is a script tool to create, "
        "run, analyse a DIRAC calculation "
        "or a buntch of calculations"
    )

    # input generation script
    subparsers = pypam_parser.add_subparsers()
    parser_input = subparsers.add_parser(
        "input", help="Tools for creating inputs for a DIRAC calculation"
    )
    parser_input.add_argument("-c", "--calc_method", help="calculation method")
    parser_input.add_argument(
        "-d", "--dirname", help="where to generate all input files for a DIRAC calculation"
    )
    parser_input.set_defaults(func=get_inp)

    # atomic database script
    parser_atomdb = subparsers.add_parser(
        "atomdb", help="Tools for compute polarizability " "from a calculation directory"
    )
    parser_atomdb.add_argument(
        "dir_list", nargs="+", type=str, help="A list of directories to compute polarizability"
    )
    parser_atomdb.add_argument(
        "-p",
        "--patterns",
        nargs="+",
        type=str,
        help="to specify which kind of directory should be considered",
    )
    parser_atomdb.add_argument(
        "--deepth",
        type=int,
        default=0,
        help="how deep of directory with respect to current directory specified "
        "by '--dirname' to find the calculation information",
    )
    # parser_atomdb.add_argument(
    #     '--verbos', nargs=1, type=int, help="print level", default=1)

    parser_atomdb.set_defaults(func=get_atomDB)
    parser_atomdb.set_defaults(patterns=["dyall"])
    parser_atomdb.set_defaults(dir_list=["./"])

    # atomic basis set
    parser_basis = subparsers.add_parser(
        "basis",
        help="Tools for get explicit basis set based on mol file " "in which default basis is used",
    )
    parser_basis.add_argument("element_type", type=str, help="Element type")
    parser_basis.add_argument(
        "basis_type",
        type=str,
        help="Basis type. At present, this script only supports "
        "dyall.acv4z, dyall.acv3z, dyall.cv3z, and dyall.cv4z",
    )
    parser_basis.add_argument(
        "-f",
        "--filename_out",
        nargs="+",
        type=str,
        help="A list of directories to compute polarizability",
    )
    parser_basis.set_defaults(func=get_explicit_bs)

    try:
        import argcomplete

        argcomplete.autocomplete(pypam_parser)
    except ImportError:
        # argcompolete not present
        pass

    args = pypam_parser.parse_args()
    # args = pypam_parser.parse_args(['input', '-d', './'])
    # args = pypam_parser.parse_args(['atomdb', './', '-p', 'faegri', 'dyall'])
    # args = pypam_parser.parse_args(['atomdb', './'])
    # print(args)

    try:
        getattr(args, "func")
    except AttributeError:
        pypam_parser.print_help()
        sys.exit(-1)
    return args.func(args)


def get_explicit_bs(args):
    fout = args.filename_out
    get_custom_basis_from_ele(args.element_type, args.basis_type, fout)


if __name__ == "__main__":
    main()



import tempfile

from pydirac import AtomicCCSet, AtomicDHFSet, Molecule

molecule = Molecule(["Mc"], [[0.0, 0.0, 0.0]])
dhf = AtomicDHFSet(
    molecule=molecule,
    is_ff=True,
    ff_mode="Q",
    hamiltonian_mode="2C",
    is_spinfree=True,
    user_mol_settings={"basis_type": "dyall.cv3z"},
)
with tempfile.TemporaryDirectory() as tmp_dir:
    dhf.write_input(output_dir=tmp_dir)
    print(" Inp file ".center(80, "#"))
    print(dhf.inp)
    print(" Mol file ".center(80, "#"))
    print(dhf.mol)

cc = AtomicCCSet(
    molecule=molecule,
    no_T=True,
    e_min=-30,
    e_max=30,
    nelec=(2, 1),
    nel_f1=(2, 1, 0, 0),
    is_spinfree=True,
    hamiltonian_mode="4C",
    is_ff=True,
    ff_mode="D",
)

# Use a temporary directory to store the output files
with tempfile.TemporaryDirectory() as tmp_dir:
    cc.write_input(output_dir=tmp_dir)
    print(" Inp file ".center(80, "#"))
    print(cc.inp)
    print(" Mol file ".center(80, "#"))
    print(cc.mol)
